#! /usr/bin/env python

import getpass
import hashlib
import json
import logging
import os
import re
import sys
import time
import urllib

from functools import wraps
from urlparse import urljoin

import requests

from requests.exceptions import ConnectionError, Timeout
from requests_oauthlib import OAuth1Session


BASE_DIR = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
DOWNLOAD_DIR = os.path.join(BASE_DIR, 'u1-files')
SSO_URL = u'https://login.ubuntu.com/'

BASE_API_URL = u'https://one.ubuntu.com/api/file_storage/v1'
BASE_FILES_URL = u'https://files.one.ubuntu.com/'
BASE_CONTENT_URL = urljoin(BASE_FILES_URL, '/content')
TRIES = 5

START_MARKER = '---- Starting download ----'
LOG_FILENAME = 'u1_downloader.log'
REPORT = {'total': 0, 'current': 0, 'errors': []}

VALID_CHARS = re.compile(r"[-_a-zA-Z0-9\. ()#@$&+~,]")
FILENAME_MAX_LENGTH = 255


logging.basicConfig(
    filename=LOG_FILENAME, level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s')


class AuthenticationError(Exception):
    """U1 authentication failed."""


class TwoFactorRequired(Exception):
    """2FA otp required."""


class ResponseNotOK(Exception):
    """An unexpected non 2xx HTTP response."""


def _encode(name):
    clean = "".join(urllib.quote(x.encode('utf-8'))
                    if not VALID_CHARS.match(x) else x for x in name)
    as_bytes = clean.encode('ascii')
    return as_bytes


def _log_error(path, error_msg):
    encoded_path = _encode(path)
    msg = "E: %s: %s" % (encoded_path, error_msg)
    logging.error(msg)
    REPORT['errors'].append(msg)
    sys.stderr.write('%s\n' % msg)


def _update_progress(size):
    if size:
        REPORT['current'] += size
        if REPORT['total'] > 0:
            status = REPORT['current'] * 100 / REPORT['total']
            print '[Overall progress: %d%%]' % status


def retry(ExceptionToCheck, tries=TRIES, delay=3, backoff=2, logger=None):
    """Retry calling the decorated function using an exponential backoff.

    Original:
    http://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/
    """
    def deco_retry(f):
        @wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except ExceptionToCheck as e:
                    msg = "%r: Retrying in %d seconds..." % (e, mdelay)

                    if logger:
                        logger.warning(msg)
                    else:
                        print msg

                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            return f(*args, **kwargs)

        return f_retry

    return deco_retry


def _make_dir(parent, dirname):
    path = os.path.join(parent, _encode(dirname))
    try:
        os.mkdir(path)
    except OSError:
        # directory already exists
        pass
    return path


def _get_token(email, password, otp=None):
    """Return SSO token details."""
    headers = {'content-type': 'application/json',
               'accept': 'application/json'}
    data = {'email': email, 'password': password,
            'token_name': 'u1-downloader'}
    if otp:
        data['otp'] = otp

    response = requests.post(
        urljoin(SSO_URL, u'/api/v2/tokens/oauth'),
        data=json.dumps(data), headers=headers)

    if response.status_code not in (200, 201):
        try:
            data = response.json()
        except ValueError:
            data = None

        logging.warning("Authentication error: %r, status_code: %s",
                        data, response.status_code)
        if data and data.get('code') == 'TWOFACTOR_REQUIRED':
            raise TwoFactorRequired()

        raise AuthenticationError("Authentication failed.")

    return response.json()


def get_session(email, password):
    """Return a U1 session to work with."""
    try:
        credentials = _get_token(email, password)
    except TwoFactorRequired:
        otp = raw_input('2FA code: ')
        credentials = _get_token(email, password, otp=otp)

    session = OAuth1Session(
        credentials.get('consumer_key'), credentials.get('consumer_secret'),
        credentials.get('token_key'), credentials.get('token_secret'),
        signature_method='PLAINTEXT')

    return session


@retry((ConnectionError, Timeout, ValueError), logger=logging)
def retryable_get(u1, url, is_json=True, stream=False):
    """Request a U1 url, retrying on network errors."""
    response = u1.get(url, stream=stream, timeout=60)

    logging.info("GET: %s, status_code: %s", url, response.status_code)
    if not response.ok:
        logging.error(response.content)
        raise ResponseNotOK()

    if is_json:
        response = response.json()

    return response


def calculate_hash(filepath):
    """Return sha1 hash as calculated by U1."""
    sha1 = hashlib.sha1()
    with open(filepath, 'rb') as f:
        while True:
            chunk = f.read(1048576)
            if not chunk:
                break
            sha1.update(chunk)
    return 'sha1:%s' % sha1.hexdigest()


def download_file(u1, node, parent):
    """Download U1 file and verify sha1 hash if provided."""
    success = True

    content_path = node['content_path']
    size = node['size']

    if size is None:
        logging.info('%s: skipped, size is null.', content_path)
        return success

    resource = _encode(node['resource_path'])
    node_id = node['key']
    filename = os.path.basename(content_path)
    encoded_filename = _encode(filename)

    print resource,

    encoded_length = len(encoded_filename)
    # check filename length after encoding
    if encoded_length > FILENAME_MAX_LENGTH:
        # use node_id
        encoded_id = _encode(node_id)
        original_name, extension = os.path.splitext(encoded_filename)
        logging.info('File name too long: %r, download as %r',
                     encoded_filename, encoded_id)
        encoded_filename = encoded_id + extension
        print '(as %s)' % encoded_filename,

    destination = os.path.join(parent, encoded_filename)

    sys.stdout.flush()
    # check if file is already downloaded
    expected_sha1 = node['hash']
    if (os.path.exists(destination) and expected_sha1):
        print '...verifying hash...',
        if (calculate_hash(destination) == expected_sha1):
            print 'OK',
            return success
        print 'failed',

    print '...downloading',
    sys.stdout.flush()
    url = urljoin(BASE_CONTENT_URL, urllib.quote(content_path.encode('utf-8')))
    try:
        response = retryable_get(u1, url, is_json=False, stream=True)
    except Exception:
        logging.exception('Download failed: %r', url)
        response = None

    if response is None:
        try:
            # retry download using node_id URL
            url = '%s%s' % (BASE_FILES_URL, node_id)
            logging.info('Retrying using node id url: %r', url)
            params = dict(next=url)
            url = ('https://one.ubuntu.com/api/1.0/from_oauth/?' +
                   urllib.urlencode(params))
            response = retryable_get(u1, url, is_json=False, stream=True)
        except Exception:
            logging.exception('Download failed: %r', url)
            return False

    sha1 = hashlib.sha1()
    with open(destination, 'wb') as f:
        for chunk in response.iter_content(chunk_size=65536):
            sha1.update(chunk)
            f.write(chunk)
        f.flush()
        os.fsync(f.fileno())

    download_sha1 = 'sha1:%s' % sha1.hexdigest()
    if expected_sha1 and download_sha1 != expected_sha1:
        logging.info("%s: expected hash: %s, actual hash: %s",
                     content_path, expected_sha1, download_sha1)
        success = False

    return success


def _get_node_metadata(u1, path):
    """Return node metadata for the given path."""
    node = None
    metadata_url = u'%s%s?include_children=true' % (
        BASE_API_URL, urllib.quote(path.encode('utf-8')))

    try:
        node = retryable_get(u1, metadata_url)
    except Exception:
        logging.exception('Error getting node information: %r', metadata_url)

    return node


def download_node(u1, path, parent):
    """Recursively download U1 node."""
    node = _get_node_metadata(u1, path)

    if node is None or (node['has_children'] and 'children' not in node):
        logging.error('Error getting node information: %r', path)
        if '?' in path:
            # if ? in path, pre-quote before retrying node information
            path = path.replace('?', urllib.quote('?'))
            node = _get_node_metadata(u1, path)

    if node is None:
        logging.error('Error getting node information: %r', path)
        _log_error(path, 'Could not get node information (%r).' % path)
        return

    try:
        node_name = os.path.basename(node['content_path'])
        dirname = _make_dir(parent, node_name)
        logging.info('%s: Starting download.', path)

        children = node['children'] if node['has_children'] else []
        for node in children:
            if node['kind'] == 'directory':
                download_node(u1, node['resource_path'], parent=dirname)

            elif node['kind'] == 'file':
                try:
                    success = download_file(u1, node, parent=dirname)
                except Exception:
                    logging.exception('Error while downloading the file.')
                    success = False
                _update_progress(node.get('size'))
                if not success:
                    _log_error(node['content_path'], 'File download failed.')

    except Exception as e:
        _log_error(path, 'Could not get node information (%r).' % e)
        logging.exception('Unexpected response, got %r', node)


def print_report():
    print '-' * 25
    errors = REPORT['errors']
    if errors:
        print '# failed downloads:', len(errors)
        for error in errors:
            print error
        print 'You can re-run this program to retry failed downloads.'
        print 'Or check FAQ: https://help.ubuntu.com/community/U1/Downloader'
    else:
        print 'All files successfully downloaded.'


def main():
    os.environ['REQUESTS_CA_BUNDLE'] = os.path.join(BASE_DIR, 'cacert.pem')

    email = raw_input('U1 email: ')
    password = getpass.getpass('U1 password: ')
    u1 = get_session(email, password)

    # get user metadata
    metadata_url = BASE_API_URL
    metadata = retryable_get(u1, metadata_url)

    # nodes to be downloaded
    nodes = [metadata['root_node_path']] + metadata['user_node_paths']
    node_names = ', '.join([_encode(n) for n in nodes])

    used_bytes = metadata['used_bytes']
    print 'Total: %s bytes' % used_bytes
    print 'Nodes: %s' % node_names
    REPORT['total'] = used_bytes
    logging.info(START_MARKER)
    logging.info('Bytes to download: %s', used_bytes)
    logging.info('Nodes: %s', node_names)

    dirname = _make_dir(BASE_DIR, 'u1-files')
    for node in nodes:
        download_node(u1, node, parent=dirname)

    # show final information
    print_report()
    getpass.getpass('Press Enter to exit...')


if __name__ == '__main__':
    main()
